import uuid
from dataclasses import dataclass
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union, cast

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    Paragraph,
    Sentence,
    UnescapedText,
)
from turnip_text.helpers import block_scope_builder, inline_scope_builder
from turnip_text.renderers import (
    CustomEmitDispatch,
    MutableState,
    Plugin,
    Renderer,
    StatelessContext,
    stateful,
    stateless,
)
from turnip_text.renderers.std_plugins import (
    CitationPluginInterface,
    CiteKey,
    FootnotePluginInterface,
    FormatPluginInterface,
    SectionPluginInterface,
)

from .base import LatexRenderer, RawLatex

CiteKeyWithOptionNote = Tuple[CiteKey, Optional[UnescapedText]]


@dataclass(frozen=True)
class FootnoteAnchor(Inline):
    label: str


@dataclass(frozen=True)
class HeadedBlock(Block):
    latex_name: str
    name: UnescapedText
    contents: BlockScope
    label: Optional[str] = None
    num: bool = True


@dataclass(frozen=True)
class Citation(Inline):
    # List of (label, note?)
    labels: List[CiteKeyWithOptionNote]


@dataclass(frozen=True)
class NamedUrl(Inline, InlineScopeBuilder):
    url: str
    name: Optional[InlineScope] = None

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        assert self.name is None
        return NamedUrl(self.url, inls)


@dataclass(frozen=True)
class DisplayList(Block):
    # TODO allow nested lists
    # items: List[Union[BlockNode, List]]
    items: List["DisplayListItem"]
    mode: str


@dataclass(frozen=True)
class DisplayListItem(Block):
    item: Block


@dataclass(frozen=True)
class Formatted(Inline):
    format_type: str  # e.g. "emph"
    items: InlineScope


class LatexCitationPlugin(Plugin[LatexRenderer], CitationPluginInterface):
    # TODO require biblatex

    _citations: Dict[str, Any]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}

    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_inline(Citation, self._emit_citation)

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[LatexRenderer], None]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._emit_bibliography),)

    def _emit_citation(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        citation: Citation,
    ) -> None:
        if any(note for _, note in citation.labels):
            # We can't add a citenote for multiple citations at a time in Latex - split individually
            for key, note in citation.labels:
                renderer.emit_macro("cite")
                if note:
                    renderer.emit_sqr_bracketed(note)
                renderer.emit_braced(str(key))
        else:
            renderer.emit_macro("cite")
            renderer.emit_braced(",".join(key for key, _ in citation.labels))

    def _emit_bibliography(self, renderer: LatexRenderer) -> None:
        renderer.emit("""{
\\raggedright
\\printbibliography
}""")

    @stateless
    def cite(
        self, ctx: StatelessContext[LatexRenderer], *labels: Union[str, Tuple[str, str]]
    ) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]
        return Citation(adapted_labels)

    # TODO make this output \citeauthor
    @stateless
    def citeauthor(self, ctx: StatelessContext[LatexRenderer], label: str) -> Inline:
        return Citation([(label, None)])


class LatexFootnotePlugin(Plugin[LatexRenderer], FootnotePluginInterface):
    _footnotes: Dict[str, Block]

    def __init__(self) -> None:
        super().__init__()

        self._footnotes = {}

    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_inline(FootnoteAnchor, self._emit_footnote_anchor)

    def _emit_footnote_anchor(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        footnote: FootnoteAnchor,
    ) -> None:
        # TODO - intelligent footnotetext placement using floats?
        renderer.emit_macro("footnote")
        renderer.emit_braced(self._footnotes[footnote.label])

    @property
    @stateful
    def footnote(self, state: MutableState[LatexRenderer]) -> InlineScopeBuilder:
        @inline_scope_builder
        def footnote_builder(contents: InlineScope) -> Inline:
            label = str(uuid.uuid4())
            self._footnotes[label] = Paragraph([Sentence([contents])])
            return FootnoteAnchor(label)

        return footnote_builder

    @stateless
    def footnote_ref(self, ctx: StatelessContext[LatexRenderer], label: str) -> Inline:
        return FootnoteAnchor(label)

    @stateful
    def footnote_text(
        self, state: MutableState[LatexRenderer], label: str
    ) -> BlockScopeBuilder:
        # Store the contents of a block scope and associate them with a specific footnote label
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Optional[Block]:
            self._footnotes[label] = contents
            return None

        return handle_block_contents


class LatexSectionPlugin(Plugin[LatexRenderer], SectionPluginInterface):
    _pagebreak_before: List[str]

    def __init__(self, pagebreak_before: List[str] = []) -> None:
        super().__init__()
        self._pagebreak_before = pagebreak_before

    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_block(HeadedBlock, self._emit_headed_block)

    def _emit_headed_block(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        block: HeadedBlock,
    ) -> None:
        if block.latex_name in self._pagebreak_before:
            renderer.emit_raw("\\pagebreak\n")
        if block.num:
            renderer.emit_macro(block.latex_name) # i.e. r"\section"
        else:
            renderer.emit_macro(block.latex_name + "*")
        renderer.emit_braced(block.name) # i.e. r"\section*" + "{Section Name}"
        if block.label:
            renderer.emit_macro("label")
            renderer.emit_braced(block.label) # i.e. r"\section*{Section Name}" + r"\label{block_label}"
        renderer.emit_break_paragraph()
        renderer.emit_blockscope(block.contents)

    @stateful
    def section(
        self,
        state: MutableState[LatexRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="section",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents

    @stateful
    def subsection(
        self,
        state: MutableState[LatexRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="subsection",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents

    @stateful
    def subsubsection(
        self,
        state: MutableState[LatexRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                latex_name="subsubsection",
                name=UnescapedText(name),
                label=label,
                num=num,
                contents=contents,
            )

        return handle_block_contents


@inline_scope_builder
def emph_builder(items: InlineScope) -> Inline:
    return Formatted("emph", items)


@inline_scope_builder
def italic_builder(items: InlineScope) -> Inline:
    return Formatted("textit", items)


@inline_scope_builder
def bold_builder(items: InlineScope) -> Inline:
    return Formatted("textbf", items)


class LatexFormatPlugin(Plugin[LatexRenderer], FormatPluginInterface):
    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_inline(Formatted, self._emit_formatted)

    def _emit_formatted(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        item: Formatted,
    ) -> None:
        renderer.emit_macro(item.format_type)
        renderer.emit_braced(item.items)

    emph = emph_builder
    italic = italic_builder
    bold = bold_builder

    OPEN_DQUOTE = RawLatex("``")
    CLOS_DQUOTE = RawLatex("''")

    @property
    @stateless
    def enquote(self, ctx: StatelessContext[LatexRenderer]) -> InlineScopeBuilder:
        @inline_scope_builder
        def enquote_builder(items: InlineScope) -> Inline:
            return InlineScope(
                [LatexFormatPlugin.OPEN_DQUOTE]
                + list(items)
                + [LatexFormatPlugin.CLOS_DQUOTE]
            )

        return enquote_builder


class LatexListPlugin(Plugin[LatexRenderer]):
    indent_list_items: bool = True

    def __init__(self, indent_list_items: bool = True):
        self.indent_list_items = indent_list_items

    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_block(DisplayList, self._emit_list)
        handler.add_custom_block(DisplayListItem, self._emit_list_item)

    def _emit_list(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        list: DisplayList,
    ) -> None:
        with renderer.emit_env(list.mode):
            renderer.emit(*list.items, joiner=renderer.emit_break_paragraph)

    def _emit_list_item(
        self,
        renderer: LatexRenderer,
        ctx: StatelessContext[LatexRenderer],
        list_item: DisplayListItem,
    ) -> None:
        renderer.emit_macro("item")
        # Put {} after \item so square brackets at the start of render_block don't get swallowed as arguments
        renderer.emit("{} ")
        indent_width = len("\\item{} ")
        with renderer.indent(indent_width):
            renderer.emit(list_item.item)

    @property
    @stateless
    def enumerate(self, ctx: StatelessContext[LatexRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def enumerate_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(
                mode="enumerate", items=cast(List[DisplayListItem], items)
            )

        return enumerate_builder

    @property
    @stateless
    def itemize(self, ctx: StatelessContext[LatexRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def itemize_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(mode="itemize", items=cast(List[DisplayListItem], items))

        return itemize_builder

    @property
    @stateless
    def item(self, ctx: StatelessContext[LatexRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def item_builder(block_scope: BlockScope) -> Block:
            return DisplayListItem(block_scope)

        return item_builder


class LatexUrlPlugin(Plugin[LatexRenderer]):
    # TODO add dependency on hyperref!!

    def _add_emitters(self, handler: CustomEmitDispatch[LatexRenderer]) -> None:
        handler.add_custom_inline(NamedUrl, self._emit_url)

    def _emit_url(
        self, renderer: LatexRenderer, ctx: StatelessContext[LatexRenderer], url: NamedUrl
    ) -> None:
        assert "}" not in url.url
        assert "#" not in url.url
        if url.name is None:
            renderer.emit_macro("url")
            renderer.emit_braced(url.url)
        else:
            renderer.emit_macro("href")
            renderer.emit_braced(url.url)
            renderer.emit_braced(url.name)

    @stateless
    def url(
        self, ctx: StatelessContext[LatexRenderer], url: str, name: Optional[str] = None
    ) -> Inline:
        return NamedUrl(
            url, name=InlineScope([UnescapedText(name)]) if name is not None else None
        )



import abc
from typing import List, Optional, Protocol, Union, runtime_checkable

__all__ = [
    "Block",
    "BlockScope",
    "Inline",
    "InlineScope",
    "Paragraph",
    "RawText",
    "Sentence",
    "UnescapedText",
    "coerce_to_block_scope",
    "coerce_to_inline_scope",
    "BlockScopeBuilder",
    "InlineScopeBuilder",
    "RawScopeBuilder",
    "CoercibleToInline",
    "CoercibleToInlineScope",
    "CoercibleToBlock",
    "CoercibleToBlockScope",
    "parse_file_native",
    "parse_str_native",
]

from ._native import (  # type: ignore
    BlockScope,
    InlineScope,
    Paragraph,
    RawText,
    Sentence,
    UnescapedText,
    coerce_to_block_scope,
    coerce_to_inline_scope,
)
from ._native import parse_file as parse_file_native
from ._native import parse_str as parse_str_native


@runtime_checkable
class Inline(Protocol):
    is_inline: bool = True


@runtime_checkable
class Block(Protocol):
    is_block: bool = True


class BlockScopeBuilder(abc.ABC):
    @abc.abstractmethod
    def build_from_blocks(self, bs: BlockScope) -> Optional[Block]:
        ...

    def __matmul__(self, maybe_b: 'CoercibleToBlockScope') -> Optional[Block]:
        bs = coerce_to_block_scope(maybe_b)
        return self.build_from_blocks(bs)


class InlineScopeBuilder(abc.ABC):
    @abc.abstractmethod
    def build_from_inlines(self, inls: InlineScope) -> Inline:
        ...

    def __matmul__(self, maybe_inls: 'CoercibleToInlineScope') -> Inline:
        inls = coerce_to_inline_scope(maybe_inls)
        return self.build_from_inlines(inls)
    
@runtime_checkable
class RawScopeBuilder(Protocol):
    def build_from_raw(self, raw: str) -> Union[Inline, Block]:
        ...


# The types that can be coerced into an Inline, in the order they are checked and attempted.
# List[Inline] is coerced by wrapping it in an InlineScope
CoercibleToInline = Union[Inline, List[Inline], str, int, float]

# The types that can be coerced into an InlineScope, in the order they are checked and attempted.
# 1. InlineScopes are passed through.
# 2. Coercion to Inline is attempted, and must succeed.
# 3. If it coerced to InlineScope by the inline process (i.e. it was originally List[Inline]),
# that InlineScope is passed through.
# 4. Otherwise the plain Inline is wrapped in InlineScope([plain_inline])
CoercibleToInlineScope = Union[InlineScope, CoercibleToInline]

# The types that can be coerced into a Block, in the order they are checked and attempted
CoercibleToBlock = Union[
    List[Block], Block, Paragraph, Sentence, CoercibleToInlineScope
]

# The types that can be coerced into a BlockScope, in the order they are checked and attempted
CoercibleToBlockScope = Union[BlockScope, CoercibleToBlock]



from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

@runtime_checkable
class Inline(Protocol):
    is_inline: bool = True


@runtime_checkable
class Block(Protocol):
    is_block: bool = True

# The types that can be coerced into an Inline, in the order they are checked and attempted.
# List[Inline] is coerced by wrapping it in an InlineScope
CoercibleToInline = Union[Inline, List[Inline], str, int, float]

# The types that can be coerced into an InlineScope, in the order they are checked and attempted.
# 1. InlineScopes are passed through.
# 2. Coercion to Inline is attempted, and must succeed.
# 3. If it coerced to InlineScope by the inline process (i.e. it was originally List[Inline]),
# that InlineScope is passed through.
# 4. Otherwise the plain Inline is wrapped in InlineScope([plain_inline])
CoercibleToInlineScope = Union[InlineScope, CoercibleToInline]

# The types that can be coerced into a Block, in the order they are checked and attempted
CoercibleToBlock = Union[
    List[Block], Block, Paragraph, Sentence, CoercibleToInlineScope
]

# The types that can be coerced into a BlockScope, in the order they are checked and attempted
CoercibleToBlockScope = Union[BlockScope, CoercibleToBlock]

def parse_file_native(path: str, locals: Dict[str, Any]) -> BlockScope: ...
def parse_str_native(data: str, locals: Dict[str, Any]) -> BlockScope: ...
def coerce_to_inline(obj: CoercibleToInline) -> Inline: ...
def coerce_to_inline_scope(obj: CoercibleToInlineScope) -> InlineScope: ...
def coerce_to_block(obj: CoercibleToBlock) -> Block: ...
def coerce_to_block_scope(obj: CoercibleToBlockScope) -> BlockScope: ...

class UnescapedText(Inline):
    def __init__(self, text: str) -> None: ...
    @property
    def text(self) -> str: ...

class RawText(Inline):
    def __init__(self, text: str) -> None: ...
    @property
    def text(self) -> str: ...

# Note - Sentence is NOT an Inline. This means there's always a hierarchy of Paragraph -> many Sentences -> many Inlines.
# InlineScopes can be nested, Sentences cannot.
class Sentence:
    def __init__(self, list: Optional[List[Inline]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the inline blocks in the sentence
    def __iter__(self) -> Iterator[Inline]: ...
    # Push an inline node into the sentence
    def push_inline(self, node: Inline) -> None: ...

class Paragraph(Block):
    def __init__(self, list: Optional[List[Sentence]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the sentences in the Paragraph
    def __iter__(self) -> Iterator[Sentence]: ...
    # Push a sentence into the Paragraph
    def push_sentence(self, s: Sentence) -> None: ...

class BlockScope(Block):
    def __init__(self, list: Optional[List[Block]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the blocks in the BlockScope
    def __iter__(self) -> Iterator[Block]: ...
    # Push a block into the BlockScope
    def push_block(self, b: Block) -> None: ...

class InlineScope(Inline):
    def __init__(self, list: Optional[List[Inline]] = None): ...
    def __len__(self) -> int: ...
    # Iterate over the inline items in the InlineScope
    def __iter__(self) -> Iterator[Inline]: ...
    # Push an inline item into the InlineScope
    def push_inline(self, b: Inline) -> None: ...



import abc
import inspect
from contextlib import contextmanager
from io import StringIO
from os import PathLike
from typing import (
    Any,
    Callable,
    Concatenate,
    Dict,
    Generator,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    ParamSpec,
    Self,
    Sequence,
    Tuple,
    Type,
    TypedDict,
    TypeVar,
)

from turnip_text import (
    Block,
    BlockScope,
    Inline,
    InlineScope,
    Paragraph,
    Sentence,
    UnescapedText,
    parse_file_native,
)

# A renderer iterates depth-first through a tree of Blocks and emits text into a buffer before writing it to a file.
# The capabilities of the renderer reflect the capabilities of the format, and different formats have different capabilities.
# Thus we need plugins to extend behaviour differently for different renderers, and to allow user-generated code.

# Plugins can have stateful or stateless functions.
# Example of a stateful plugin: a TODO list which tracks and re-states the TODOs for a file at the end, anything that creates new labels.
# Example of a stateless plugin: inline formatting, simple shortcut macros
# In either case, the plugin may want to create new renderable items, and thus may be renderer-specific - a bibliography may use a LaTeX bibtex backend or manually compute citation text for Markdown, and an inline formatter will use different Latex or Markdown primitives for formatting.

# It can be useful to use stateless functions *while* rendering, once the document layout has frozen.
# e.g. when rendering a TODO item, using `ctx.bold @ ctx.color(red) @ item_text` is more convenient then directly emitting format primitives.
# This is safe because those plugins are stateless, but e.g. creating a label inside the render function may result in unexpected behaviour.

TRenderer = TypeVar("TRenderer", bound="Renderer")

TBlock = TypeVar("TBlock", bound=Block)
TInline = TypeVar("TInline", bound=Inline)


class CustomEmitDispatch(Generic[TRenderer]):
    # mypy doesn't let us use TBlock/TInline here - they're not bound to anything
    _block_table: Dict[
        Type[Block], Callable[[TRenderer, "StatelessContext[TRenderer]", Block], None]
    ]
    _inline_table: Dict[
        Type[Inline], Callable[[TRenderer, "StatelessContext[TRenderer]", Inline], None]
    ]

    def __init__(self) -> None:
        super().__init__()
        self._block_table = {}
        self._inline_table = {}

    def add_custom_block(
        self,
        t: Type[TBlock],
        f: Callable[[TRenderer, "StatelessContext[TRenderer]", TBlock], None],
    ) -> None:
        if t in self._block_table:
            raise RuntimeError(f"Conflict: registered two renderers for {t}")
        # We know that we only assign _block_table[t] if f takes t, and that when we pull it
        # out we will always call f with something of type t.
        # mypy doesn't know that, so we say _block_table stores functions taking Block (the base class)
        # and sweep the difference under the rug
        self._block_table[t] = f # type: ignore

    def add_custom_inline(
        self,
        t: Type[TInline],
        f: Callable[[TRenderer, "StatelessContext[TRenderer]", TInline], None],
    ) -> None:
        if t in self._inline_table:
            raise RuntimeError(f"Conflict: registered two renderers for {t}")
        # as above
        self._inline_table[t] = f # type: ignore

    def emit_block(
        self, renderer: TRenderer, ctx: "StatelessContext[TRenderer]", obj: TBlock
    ) -> None:
        f = self._block_table.get(type(obj))
        if f is None:
            for t, f in self._block_table.items():
                if isinstance(obj, t):
                    f(renderer, ctx, obj)
                    return
            raise NotImplementedError(f"Couldn't handle {obj}")
        else:
            f(renderer, ctx, obj)

    def emit_inline(
        self, renderer: TRenderer, ctx: "StatelessContext[TRenderer]", obj: TInline
    ) -> None:
        f = self._inline_table.get(type(obj))
        if f is None:
            for t, f in self._inline_table.items():
                if isinstance(obj, t):
                    f(renderer, ctx, obj)
                    return
            raise NotImplementedError(f"Couldn't handle {obj}")
        else:
            f(renderer, ctx, obj)


TPlugin = TypeVar("TPlugin", bound="Plugin[Any]")
P = ParamSpec("P")
T = TypeVar("T")


class Plugin(Generic[TRenderer]):
    # Initialized when the plugin is included into the MutableState.
    # Should always be non-None when the plugin's emitted functions are called
    __state: "MutableState[TRenderer]" = None  # type: ignore
    __ctx: "StatelessContext[TRenderer]" = None  # type: ignore

    def __init_ctx(
        self, state: "MutableState[TRenderer]", ctx: "StatelessContext[TRenderer]"
    ) -> None:
        assert self.__state is None and self.__ctx is None
        self.__state = state
        self.__ctx = ctx

    @property
    def _plugin_name(self) -> str:
        return type(self).__name__

    def _interface(self) -> Dict[str, Any]:
        """
        Define the interface available to the renderer context,
        and thus all eval-brackets in evaluated documents.

        By default, finds all public variables, properties (as `property` objects, without evaluating them),member functions, and static functions.
        Ignores any fields that begin with _.
        Based on https://github.com/python/cpython/blob/a0773b89dfe5cd2190d539905dd89e7f6455668e/Lib/inspect.py#L562C5-L562C5.

        May be overridden."""

        interface = {}
        names = dir(self)
        # Ignore "DynamicClassAttributes" for now, `self` isn't a class so doesn't have them
        for key in names:
            if key.startswith("_"):
                continue
            value = inspect.getattr_static(self, key)
            if isinstance(value, property):
                # Hack to make @property @stateless work.
                stateless = getattr(value.fget, "_stateless", False)
                interface[key] = BoundProperty(self, value, stateless)
            elif inspect.ismethoddescriptor(value) or inspect.isdatadescriptor(value):
                # We want to pass these through to the state/less context objects directly so they are re-__get__ted each time, but they still need to be bound to this plugin object.
                # The solution: construct BoundProperty, which binds the internal __get__, __set__, __delete__ to self,
                # and pass that through as the interface.
                stateless = getattr(value, "_stateless", False)
                interface[key] = BoundProperty(self, value, stateless)
            else:
                # Use getattr to do everything else as normal e.g. bind methods to the object etc.
                interface[key] = getattr(self, key)

        return interface

    def _add_emitters(self, handler: CustomEmitDispatch[TRenderer]) -> None:
        """
        Add render handler functions for all custom Blocks and Inlines this plugin uses
        """
        ...

    # TODO improve/remove amble handlers - these are better suited with custom blocks

    def _preamble_handlers(self) -> Iterable[Tuple[str, Callable[[TRenderer], None]]]:
        return ()

    def _postamble_handlers(self) -> Iterable[Tuple[str, Callable[[TRenderer], None]]]:
        return ()

    @staticmethod
    def _make_contexts(
        plugins: Sequence["Plugin[TRenderer]"],
    ) -> Tuple["StatelessContext[TRenderer]", "MutableState[TRenderer]"]:
        ctx: "StatelessContext[TRenderer]" = StatelessContext()
        state: "MutableState[TRenderer]" = MutableState()
        for plugin in plugins:
            i = plugin._interface()

            # Everything may or may not be stateful, therefore the stateful context can include anything
            state.__dict__.update(i)

            for key, value in i.items():
                # The stateless context only includes:
                # - properties or other data descriptors, functions, and methods that have been explicitly marked stateless.
                # - class variables which aren't data descriptors (they don't have access to the plugin self and don't have access to __state.)

                is_stateless = False
                if inspect.ismethod(value):
                    is_stateless = getattr(value.__func__, "_stateless", False)
                elif hasattr(value, "__call__") or hasattr(value, "__get__"):
                    # Callables or data descriptors need to be explicitly marked as stateless
                    is_stateless = getattr(value, "_stateless", False)
                else:
                    # It's not a bound method, it's not callable or gettable, it's probably a plain variable.
                    # It can't get access to __state, so expose it.
                    is_stateless = True

                if is_stateless:
                    ctx.__dict__[key] = value
            plugin.__init_ctx(state, ctx)
        return ctx, state

    @staticmethod
    def _stateful(
        f: Callable[Concatenate[TPlugin, "MutableState[TRenderer]", P], T]
    ) -> Callable[Concatenate[TPlugin, P], T]:
        """
        An annotation for plugin bound methods which access the __state object i.e. other stateful (and stateless) functions and variables.
        This is the only way to access the state, and thus theoretically the only way to mutate state.
        Unfortunately, we can't protect a plugin from modifying its private state in a so-annotated "stateless" function.
        """

        def wrapper(plugin: TPlugin, /, *args: Any, **kwargs: Any) -> T:
            if plugin.__state._frozen:
                raise RuntimeError(
                    "Can't run a stateful function when the state is frozen!"
                )
            return f(plugin, plugin.__state, *args, **kwargs)

        wrapper._stateful = True  # type: ignore
        wrapper.__doc__ = f.__doc__

        return wrapper

    @staticmethod
    def _stateful_property(
        f: Callable[Concatenate[TPlugin, "MutableState[TRenderer]", P], T]
    ) -> property:
        """
        An example of stateful property:

        @stateful_property
        def todo(self, state: MutableState) -> InlineScopeBuilder:
            @inline_scope_builder
            def something_that_mutates_state_by_adding_a_todo(...):
                ...
            return something_that_mutates_state_by_adding_a_todo

        state is always checked to not be frozen when the property is accessed.

        In this example, the lambda which captures __state may live longer until state is frozen.
        It's your responsibility to make sure it doesn't mutate __state in that case!!

        Note that because property objects have opaque typing, typecheckers can't tell that this
        returns something of type T. This means if you try to use stateful_property to decorate a function
        that must match some property defined elsewhere, it won't work.
        In those cases, you can use @property above @stateful and it all works fine.
        """

        return property(Plugin._stateful(f))

    @staticmethod
    def _stateless(
        f: Callable[Concatenate[TPlugin, "StatelessContext[TRenderer]", P], T]
    ) -> Callable[Concatenate[TPlugin, P], T]:
        """
        An annotation for plugin bound methods which access the __ctx object i.e. other stateless functions.
        This is the only way to access ctx.
        Unfortunately, we can't protect a plugin from modifying its private state in a so-annotated "stateless" function.

        class SomePlugin(Plugin):
            @stateless
            def some_stateless_function(self, ctx, other_arg):
                return (ctx.bold @ other_arg)

        some_plugin.some_stateless_function(other_arg) # returns (ctx.bold @ other_arg), don't need to pass in ctx!

        TODO could make this pass through stuff and just set _stateless, if __ctx is changed to _ctx
        """

        def wrapper(plugin: TPlugin, /, *args: Any, **kwargs: Any) -> T:
            return f(plugin, plugin.__ctx, *args, **kwargs)

        wrapper._stateless = True  # type: ignore
        wrapper.__doc__ = f.__doc__

        return wrapper

    @staticmethod
    def _stateless_property(
        f: Callable[Concatenate[TPlugin, "StatelessContext[TRenderer]", P], T]
    ) -> property:
        """
        In case you want to use the StatelessContext in a property:

        @stateless_property
        def bold_or_italic(self, ctx):
            if random.randint(0, 1) == 1:
                return ctx.bold
            else:
                return ctx.italic

        Note that because property objects have opaque typing, typecheckers can't tell that this
        returns something of type T. This means if you try to use stateless_property to decorate a function
        that must match some property defined elsewhere, it won't work.
        In those cases, you can use @property above @stateless and it all works fine.
        """

        return property(Plugin._stateless(f))


stateful = Plugin._stateful
stateful_property = Plugin._stateful_property
stateless = Plugin._stateless
stateless_property = Plugin._stateless_property


class BoundProperty:
    "Emulate PyProperty_Type() in Objects/descrobject.c, as seen in https://docs.python.org/3.8/howto/descriptor.html"

    _obj_for_data_descriptor: Any
    _data_descriptor: Any
    _stateless: bool

    def __init__(self, obj: Any, data_descriptor: Any, stateless: bool) -> None:
        self._obj_for_data_descriptor = obj
        self._data_descriptor = data_descriptor
        self._stateless = stateless
        self.__doc__ = data_descriptor.__doc__

    def __get__(self, _obj: Any, _ownerclass: Optional[type] = None) -> Any:
        return self._data_descriptor.__get__(self._obj_for_data_descriptor)

    def __set__(self, _obj: Any, value: Any) -> None:
        self._data_descriptor.__set__(self._obj_for_data_descriptor, value)

    def __delete__(self, _obj: Any) -> None:
        self._data_descriptor.__delete__(self._obj_for_data_descriptor)


class StatelessContext(Generic[TRenderer]):
    def __getattr__(self, name: str) -> Any:
        # The StatelessContext has various things that we don't know at type-time.
        # We want to be able to use those things from Python code.
        # __getattr__ is called when an attribute access is attempted and the "normal routes" fail.
        # Defining __getattr__ tells the type checker "hey, this object has various dynamic attributes"
        # and the type checker will then allow Python code to use arbitrary attributes on this object.
        # We still need to raise AttributeError here, because we don't actually define any attributes this way.
        raise AttributeError(name=name, obj=self)


class MutableState(Generic[TRenderer]):
    _frozen: bool = False  # Set to True when rendering the document, which disables functions annotated with @stateful.

    def parse_file(self, path: "PathLike[Any]") -> BlockScope:
        return parse_file_native(str(path), self.__dict__)
    
    def __getattr__(self, name: str) -> Any:
        # The StatelessContext has various things that we don't know at type-time.
        # We want to be able to use those things from Python code.
        # __getattr__ is called when an attribute access is attempted and the "normal routes" fail.
        # Defining __getattr__ tells the type checker "hey, this object has various dynamic attributes"
        # and the type checker will then allow Python code to use arbitrary attributes on this object.
        # We still need to raise AttributeError here, because we don't actually define any attributes this way.
        raise AttributeError(name=name, obj=self)


# TODO Make preamble/postamble return Blocks to be rendered instead of just str? Would allow e.g. a Bibliography section? Perhaps better to expose a bibliography block for "standard" postambles?
class AmbleMap(Generic[TRenderer]):
    """Class that stores and reorders {pre,post}amble handlers.

    Handlers are simply functions that return a string, which have a unique ID.

    Only one handler per ID can exist.

    The user can request that the ID are sorted in a particular order, or default to order of insertion.

    When the document is rendered, the handlers will be called in that order."""

    _handlers: Dict[str, Callable[[TRenderer], None]]
    _id_order: List[str]

    def __init__(self) -> None:
        self._handlers = {}
        self._id_order = []

    def push_handler(self, id: str, f: Callable[[TRenderer], None]) -> None:
        if id in self._handlers:
            raise RuntimeError(f"Conflict: registered two amble-handlers for ID {id}")
        self._handlers[id] = f
        self._id_order.append(id)

    def reorder_handlers(self, selected_id_order: List[str]) -> None:
        """Request that certain handler IDs are rendered in a specific order.

        Does not need to be a complete ordering, i.e. if handlers ['a', 'b', 'c'] are registered
        this function can be called with ['c', 'a'] to ensure 'c' comes before 'a',
        but all of the IDs in the order need to have been registered.

        When the requested ordering is incomplete, handlers which haven't been mentioned
        will retain their old order but there is no specified ordering between (mentioned) and (not-mentioned) IDs.
        """

        assert all(id in self._handlers for id in selected_id_order)

        if len(selected_id_order) != len(set(selected_id_order)):
            raise RuntimeError(
                f"reorder_handlers() called with ordering with duplicate IDs: {selected_id_order}"
            )

        # Shortcut if the selected order is complete i.e. covers all IDs so far
        if len(self._id_order) == len(selected_id_order):
            self._id_order = selected_id_order
        else:
            # Otherwise, we need to consider the non-selected IDs too.
            # The easy way: put selected ones first, then non-selected ones last
            # Get the list of ids NOT in selected_id_order, in the order they're currently in in self._id_order
            non_selected_ids = [
                id for id in self._id_order if id not in selected_id_order
            ]
            self._id_order = selected_id_order + non_selected_ids

        assert all(id in self._id_order for id in self._handlers.keys())

    def generate_ambles(self, renderer: TRenderer) -> Generator[Callable[[TRenderer], None], None, None]:
        for id in self._id_order:
            yield self._handlers[id]


class Renderer(abc.ABC):
    plugins: List[Plugin[Self]]
    _ctx: StatelessContext[Self]
    _state: MutableState[Self]

    emit_dispatch: CustomEmitDispatch[Self]
    preamble_handlers: AmbleMap[Self]
    postamble_handlers: AmbleMap[Self]

    document: StringIO

    _indent: str = ""
    # After emitting a newline with emit_newline, this is set.
    # The next call to emit_raw will emit _indent.
    # This is important if you want to change the indent after something has already emitted a newline,
    # e.g. if you wrap emit_paragraph() in indent(4), the emit_paragraph() will emit a final newline but *not* immediately emit the indent of 4, so subsequent emissions are nicely indented.
    # In the same way, if you emit a newline *then* change the indent, the next emitted item will have the new indent applied.
    _need_indent: bool = False

    def __init__(self: Self, plugins: Sequence[Plugin[Self]]) -> None:
        super().__init__()

        # Create render handlers and pre/postamble handlers
        self.emit_dispatch = CustomEmitDispatch()
        self.emit_dispatch.add_custom_block(
            BlockScope, lambda r, ctx, bs: r.emit_blockscope(bs)
        )
        self.emit_dispatch.add_custom_block(
            Paragraph, lambda r, ctx, bs: r.emit_paragraph(bs)
        )
        self.emit_dispatch.add_custom_inline(
            InlineScope, lambda r, ctx, inls: r.emit_inlinescope(inls)
        )
        self.emit_dispatch.add_custom_inline(
            UnescapedText, lambda r, ctx, t: r.emit_unescapedtext(t)
        )

        self.preamble_handlers = AmbleMap()
        self.postamble_handlers = AmbleMap()

        self.plugins = list(plugins)
        self._ctx, self._state = Plugin._make_contexts(self.plugins)
        for p in self.plugins:
            p._add_emitters(self.emit_dispatch)
            for preamble_id, preamble_func in p._preamble_handlers():
                self.preamble_handlers.push_handler(preamble_id, preamble_func)
            for postamble_id, postamble_func in p._postamble_handlers():
                self.postamble_handlers.push_handler(postamble_id, postamble_func)

        self.document = StringIO()

    def request_preamble_order(self, preamble_id_order: List[str]) -> None:
        self.preamble_handlers.reorder_handlers(preamble_id_order)

    def request_postamble_order(self, postamble_id_order: List[str]) -> None:
        self.postamble_handlers.reorder_handlers(postamble_id_order)

    def parse_file(self, p: "PathLike[Any]") -> BlockScope:
        return self._state.parse_file(p)

    def emit_raw(self, x: str) -> None:
        """
        The function on which all emitters are based.
        """
        if self._need_indent:
            self.document.write(self._indent)
            self._need_indent = False
        self.document.write(x)

    def emit_newline(self) -> None:
        self.document.write("\n")
        self._need_indent = True

    # TODO pass a generator instead of emit_t, ts!
    def emit_join(self, emit_t: Callable[[T], None], ts: Iterable[T], emit_join: Callable[[], None]) -> None:
        first = True
        for t in ts:
            if not first:
                emit_join()
            first = False
            emit_t(t)

    def emit_join_gen(self, emit_gen: Generator[None, None, None], emit_join: Callable[[], None]) -> None:
        first = True
        while True:
            if not first:
                emit_join()
            first = False
            try:
                next(emit_gen)
            except StopIteration:
                break
    
    def emit_break_sentence(self) -> None:
        self.emit_newline()
    
    def emit_break_paragraph(self) -> None:
        self.emit_newline()
        self.emit_newline()

    @abc.abstractmethod
    def emit_unescapedtext(self, t: UnescapedText) -> None:
        """
        Given some text, emit a string that will look like that text exactly in the given backend.
        """
        raise NotImplementedError(f"Need to implement emit_unescapedtext")

    def render_doc(self: Self, doc_block: BlockScope) -> StringIO:
        for emit_preamble in self.preamble_handlers.generate_ambles(self):
            emit_preamble(self)
            self.emit_break_paragraph()
        self.emit_blockscope(doc_block)
        for emit_postamble in self.postamble_handlers.generate_ambles(self):
            self.emit_break_paragraph()
            emit_postamble(self)
        return self.document

    # TODO this is probably a bad idea to implement because it will get mixed up with raw.
    # def emit(self, x: Any) -> None:
    #     if isinstance(x, Inline):
    #         self.emit_inline(x)
    #     else:
    #         self.emit_block(x)


    # TODO or i could get even crazier with it - make it expand tuples?
    def emit(self, *args: Any, joiner: Optional[Callable[[], None]]=None) -> None:
        first = True            
        for a in args:
            if joiner and not first:
                joiner()
            first = False
            if isinstance(a, str):
                self.emit_raw(a)
            elif isinstance(a, Inline):
                self.emit_inline(a)
            else:
                self.emit_block(a)

    def emit_inline(self: Self, i: Inline) -> None:
        self.emit_dispatch.emit_inline(self, self._ctx, i)

    def emit_block(self: Self, b: Block) -> None:
        self.emit_dispatch.emit_block(self, self._ctx, b)

    def emit_blockscope(self, bs: BlockScope) -> None:
        # Default: join paragraphs with self.PARAGRAPH_SEP
        # If you get nested blockscopes, this will still be fine - you won't get double separators
        self.emit_join(self.emit_block, bs, self.emit_break_paragraph)

    def emit_paragraph(self, p: Paragraph) -> None:
        # Default: join sentences with self.SENTENCE_SEP
        for s in p:
            self.emit_sentence(s)

    def emit_inlinescope(self, inls: InlineScope) -> None:
        # Default: join internal inline elements directly
        for i in inls:
            self.emit_inline(i)

    def emit_sentence(self, s: Sentence) -> None:
        # Default: join internal inline elements directly
        # TODO could be extended by e.g. latex to ensure you get sentence-break-whitespace at the end of each sentence?
        for i in s:
            self.emit_inline(i)
        self.emit_break_sentence()

    def push_indent(self, n: int) -> None:
        self._indent += " " * n

    def pop_indent(self, n: int) -> None:
        if len(self._indent) < n:
            raise ValueError()
        self._indent = self._indent[:-n]

    @contextmanager
    def indent(self, n: int) -> Iterator[None]:
        self.push_indent(n)
        try: 
            yield
        finally:
            self.pop_indent(n)
# class DocumentConfig(Generic[TRenderer], abc.ABC):
#    pass

# We want to be able to
# - add custom args to Renderer.__init__ (so we can't hardcode how Renderers are created)
# - add custom reproducible transformations to a parsed document BlockScope (to make up for a lack of ambles)
# - set a default set of plugins for documents (so we have to start all this through the DocumentConfig)


from typing import Optional, Protocol, Tuple, Union

from turnip_text import BlockScopeBuilder, Inline, InlineScopeBuilder

CiteKey = str
CiteKeyWithNote = Tuple[CiteKey, str]


class CitationPluginInterface(Protocol):
    """Standard plugin for citations"""

    _BIBLIOGRAPHY_POSTAMBLE_ID: str = "CitationPluginInterface_Bibliography"
    """Standard ID for the bibliography postamble"""

    def cite(self, *keys: Union[CiteKey, CiteKeyWithNote]) -> Inline:
        """Insert an inline citation for the references with the given keys, appending notes for each one"""
        ...

    def citeauthor(self, key: CiteKey) -> Inline:
        """Insert the name of the author(s) for the reference with the given key"""
        ...


class FootnotePluginInterface(Protocol):
    """Standard plugin for footnotes"""

    @property
    def footnote(self) -> InlineScopeBuilder:
        """Define an unlabelled footnote, capturing the following inline text as the contents,
        and insert a reference to that footnote here.

        `[footnote]{contents}` is roughly equivalent to `[footnote_ref(<some_label>)]` and `[footnote_text(<some_label>)]{contents}`
        """
        ...

    def footnote_ref(self, label: str) -> Inline:
        """Insert a reference to a footnote, which will have the contents defined by a matching call to footnote_text.
        If, at render time, footnote_text has not been called with a matching label then an error will be raised.
        """
        ...

    def footnote_text(self, label: str) -> BlockScopeBuilder:
        """Define the contents of a footnote with the given label. Cannot be called twice with the same label."""
        ...


class SectionPluginInterface(Protocol):
    def section(
        self, name: str, label: Optional[str] = None, numbered: bool = False
    ) -> BlockScopeBuilder:
        ...

    def subsection(
        self, name: str, label: Optional[str] = None, numbered: bool = False
    ) -> BlockScopeBuilder:
        ...

    def subsubsection(
        self, name: str, label: Optional[str] = None, numbered: bool = False
    ) -> BlockScopeBuilder:
        ...


class FormatPluginInterface(Protocol):
    emph: InlineScopeBuilder
    italic: InlineScopeBuilder
    bold: InlineScopeBuilder

    @property
    def enquote(self) -> InlineScopeBuilder:
        ...


class ListPluginInterface(Protocol):
    @property
    def enumerate(self) -> BlockScopeBuilder:
        ...

    @property
    def itemize(self) -> BlockScopeBuilder:
        ...

    @property
    def item(self) -> BlockScopeBuilder:
        ...


class UrlPluginInterface(Protocol):
    def url(self, url: str) -> Inline:
        ...



import html
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator, List

from turnip_text import Block, BlockScope, Inline, Paragraph, UnescapedText
from turnip_text.renderers import Plugin, Renderer


@dataclass(frozen=True)
class RawMarkdown(Inline):
    md: str


class MarkdownRenderer(Renderer):
    html_mode_stack: List[bool]

    def __init__(self, plugins: List[Plugin["MarkdownRenderer"]], html_mode_only: bool=False) -> None:
        super().__init__(plugins)
        self.emit_dispatch.add_custom_inline(
            RawMarkdown, lambda r, ctx, raw: r._emit_raw_markdown(raw)
        )
        # Once you're in HTML mode, you can't drop down to Markdown mode again.
        # If they asked for HTML mode only, just make that the first entry in the stack.
        # If they didn't, we start in Markdown mode.
        self.html_mode_stack = [html_mode_only]

    def _emit_raw_markdown(self, r: RawMarkdown) -> None:
        self.emit_raw(r.md)

    def emit_unescapedtext(self, t: UnescapedText) -> None:
        if self.in_html_mode:
            self.emit_raw(html.escape(t.text))
        else:
            # note - right now this assumes we're using a unicode-compatible setup and thus don't need to escape unicode characters.
            # note - order is important here because the subsitutions may introduce more special characters.
            # e.g. if the backslash replacement applied later, the backslash from escaping "(" would be escaped as well

            # TODO - some of these are overzealous, e.g. () and -, because in most contexts they're interpreted as normal text.
            # context sensitive escaping?
            # https://www.markdownguide.org/basic-syntax/
            special_chars = r"\\`*_{}[]<>()#+-.!|"

            data = t.text
            for char in special_chars:
                data = data.replace(char, "\\" + char)
            self.emit_raw(data)

    def emit_paragraph(self, p: Paragraph) -> None:
        if self.in_html_mode:
            self.emit_raw("<p>")
            self.emit_newline()
            with self.indent(4):
                super().emit_paragraph(p)
                # emit_paragraph already ends with a newline
            self.emit_raw("</p>")
        else:
            super().emit_paragraph(p)

    # TODO do this for blocks instead of paragraphs?
    # def emit_block(self, b: Block) -> None:
    #     if self.in_html_mode:
    #         self.emit_raw("<p>")
    #         with self.indent(4):
    #             self.emit_line_break()
    #             super().emit_block(b)
    #         self.emit_line_break()
    #         self.emit_raw("</p>")
    #     else:
    #         super().emit_block(b)

    @property
    def in_html_mode(self) -> bool:
        return self.html_mode_stack[-1]
    
    @contextmanager
    def html_mode(self) -> Iterator[None]:
        self.html_mode_stack.append(True)

        try:
            yield
        finally:
            self.html_mode_stack.pop()



"""
Labels and refs galore!

We want labels to refer to content, not just points in the text.
The content it refers to is one of a bunch of types:
- floats, like figures
- a section or other structural point
- text
- footnotes?
- citation marks?
- any custom types that may or may not be floats: code listings, tables, equations, e.g. my hypotheses for the mphil diss

References to specific text are a special case: the user needs to choose how to refer back to them, maybe by closest-structural-counter, maybe they're in some float (is there a floating quote environment?), maybe by page...
It gets weird there too because some formats don't have pages. Markdown would probably change "on page X" to a link "here"?.

These "label kinds" determine the in-text references to each item, and might (per-renderer) determine the kind of \label used.
We need to find a way to make it work kinda like LaTeX - LaTeX lets you associated labels with multiple counters, and choose when counters reset, so you can say "figure 1.2.4" is figure 4 in chapter 1 in section 2.
This system also needs to support creating new sub-labels: e.g. if I already have a "figure" label, I want to create a sub-figure to represent "figure 1.2.4.a"

Clearly we need to make a hierarchy of some kind. 
Represent it as an acyclic tree.
If a LaTeX renderer wants, it could choose to implement that through configuring counters in the preamble

Dad recommends a separate Counting phase once the document has been constructed and before rendered.
This requires me to make Blocks and Inlines implement "here are my children" so I have a consistent way to DFS it and then run counting.
"""

import abc
import dataclasses
from collections import defaultdict
from dataclasses import dataclass
from typing import Dict, List, Mapping, MutableMapping, Optional, Sequence, Tuple

from turnip_text import (
    Inline,
    InlineScope,
    InlineScopeBuilder,
    UnescapedText,
    coerce_to_inline_scope,
)
from turnip_text.renderers import StatelessContext

# e.g. arabic, roman numerals (lower and upper), alphabetic
Numbering = Mapping[int, UnescapedText]

@dataclass
class LabelKind(abc.ABC):
    id: str
    sublabels: List['LabelKind']
    numbering: Numbering

    @abc.abstractmethod
    def default_backref_text(self, ctx: StatelessContext, numbers: Sequence[Tuple[Numbering, int]]) -> Inline:
        ...

@dataclass
class BasicLabelKind(LabelKind):
    prefix: str

    def default_backref_text(self, ctx: StatelessContext, numbers: Sequence[Tuple[Numbering, int]]) -> Inline:
        return coerce_to_inline_scope(
            f"{self.prefix} {'.'.join(numbering[value] for numbering, value in numbers)}"
        )

class LabelKindMap:
    # The roots of the tree of labels
    label_kind_tree: List[LabelKind]
    # Mapping of Label.id to the Label in the tree of labels
    label_kind_lookup: Mapping[str, LabelKind]

    def __init__(self, label_tree: List[LabelKind]) -> None:
        self.label_kind_tree = label_tree
        self.label_kind_lookup = {}
        
        raise NotImplementedError()
    

# Unlike the phd_notes lib version, this shouldn't be subclassed.
@dataclass(frozen=True)
class Label:
    kind: str
    id: Optional[str] # If the id is None, you can't backreference this object. Ever.
    number: Tuple[int, ...] # The numbers for all label kinds that parent this kind, and the number for this label. e.g. if figures are numbered per chapter per section, this field is (chapter no., section no., figure no.). All zero-indexed.

    def canonical(self) -> str:
        return f"{self.kind}:{self.id}"
    
    def __str__(self) -> str:
        return self.canonical()

@dataclass(frozen=True)
class Backref(Inline, InlineScopeBuilder):
    id: str
    kind: Optional[str] # Usually there should be exactly one Label for every one ID. This is used to disambiguate otherwise
    label_contents: Optional[Inline] # Override for label

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        assert self.label_contents is None
        return dataclasses.replace(self, label_contents=inls)

# Responsible for keeping track of all the labels in a document
class DocumentLabelRefState:
    _label_kind_map: LabelKindMap
    _label_kind_count: Dict[str, int]
    _label_id_to_possible_kinds: Dict[str, Dict[str, Label]]

    def __init__(self, label_kind_map: LabelKindMap) -> None:
        self._label_kind_map = label_kind_map
        self._label_kind_count = defaultdict(lambda: 0)
        self._label_id_to_possible_kinds = defaultdict(dict)

    # TODO make this need to be called with unfrozen state
    def register_new_label(self, kind: str, id: Optional[str]) -> Label:
        # TODO this numbering shouldn't be done here, it should be done in a separate pass.
        # For now it's fine, but if we later implement e.g. float reordering for different output formats
        # then it would be wrong.
        number = self._label_kind_count[kind]
        self._label_kind_count[kind] = number + 1
        l = Label(
            kind=kind,
            id=id,
            # TODO how the hell do we track the upper numbers. We need a counter class or something
            number=number
        )
        if id is not None:
            self._label_id_to_possible_kinds[id][kind] = l
        return l
    
    # TODO make this need to be called with frozen state
    def lookup_backref(self, backref: Backref) -> Label:
        if backref.id not in self._label_id_to_possible_kinds:
            raise ValueError(f"Backref {backref} refers to an ID '{backref.id}' with no label!")
        
        possible_kinds = self._label_id_to_possible_kinds[backref.id]

        if backref.kind is None:
            if len(possible_kinds) != 1:
                raise ValueError(f"Backref {backref} doesn't specify the kind of label it's referring to, and there are multiple with that ID: {possible_kinds}")
            only_possible_label = next(iter(possible_kinds.values()))
            return only_possible_label
        else:
            if backref.kind not in possible_kinds:
                raise ValueError(f"Backref {backref} specifies an label of kind {backref.kind}, which doesn't exist for ID {backref.id}: {possible_kinds}")
            return possible_kinds[backref.kind]
        
    # TODO make this need to be called with frozen state
    def resolve_backref_text(self, backref: Backref) -> Inline:
        label = self.lookup_backref(backref)



import uuid
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    Paragraph,
    Sentence,
    UnescapedText,
)
from turnip_text.helpers import block_scope_builder, inline_scope_builder
from turnip_text.renderers import (
    CustomEmitDispatch,
    MutableState,
    Plugin,
    Renderer,
    StatelessContext,
    stateful,
    stateless,
)
from turnip_text.renderers.markdown.base import MarkdownRenderer, RawMarkdown
from turnip_text.renderers.std_plugins import (
    CitationPluginInterface,
    CiteKey,
    FootnotePluginInterface,
    FormatPluginInterface,
    SectionPluginInterface,
)

CiteKeyWithOptionNote = Tuple[CiteKey, Optional[UnescapedText]]


@dataclass(frozen=True)
class FootnoteAnchor(Inline):
    label: str


@dataclass(frozen=True)
class HeadedBlock(Block):
    level: int # One-indexed
    name: UnescapedText
    contents: BlockScope
    # Numbering
    use_num: bool  # Whether to number the heading or not
    num: Tuple[int, ...]  # The actual number of the heading
    # Labelling
    label: Optional[str] = None


@dataclass(frozen=True)
class Citation(Inline):
    # List of (label, note?)
    labels: List[CiteKeyWithOptionNote]


@dataclass(frozen=True)
class NamedUrl(Inline, InlineScopeBuilder):
    url: str
    name: Optional[InlineScope] = None

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        assert self.name is None
        return NamedUrl(self.url, inls)


@dataclass(frozen=True)
class DisplayList(Block):
    # TODO allow nested lists
    # items: List[Union[BlockNode, List]]
    items: List["DisplayListItem"]
    numbered: bool


@dataclass(frozen=True)
class DisplayListItem(Block):
    contents: Block


@dataclass(frozen=True)
class Formatted(Inline):
    markdown_format: str  # e.g. "*", "**"
    html_format: str  # e.g. "b", "i"
    items: InlineScope


class MarkdownCitationAsFootnotePlugin(
    Plugin[MarkdownRenderer], CitationPluginInterface
):
    _citations: Dict[str, Any]
    _referenced_citations: Set[str]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}
        self._referenced_citations = set()

    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_inline(Citation, self._emit_citation)

    def _postamble_handlers(
        self,
    ) -> Iterable[Tuple[str, Callable[[MarkdownRenderer], None]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._emit_bibliography),)

    def _emit_citation(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        citation: Citation,
    ) -> None:
        # TODO what happens with unmarkdownable labels? e.g. labels with backslash or something. need to check that when loading.
        # TODO also maybe people wouldn't want those labels being exposed?
        assert not renderer.in_html_mode

        for label, opt_note in citation.labels:
            if opt_note is None:
                renderer.emit_raw(f"[^{label}]")
            else:
                renderer.emit(
                    f"\\([^{label}], ",
                    opt_note,
                    f"\\)"
                )

    def _emit_bibliography(self, renderer: MarkdownRenderer) -> None:
        # TODO actual reference rendering!
        def bib_gen() -> Generator[None, None, None]:
            for label in self._referenced_citations:
                renderer.emit(f"[^{label}]: ", UnescapedText(f"cite {label}"))
                yield

        renderer.emit_join_gen(bib_gen(), renderer.emit_break_paragraph)

    @stateful
    def cite(
        self,
        state: MutableState[MarkdownRenderer],
        *labels: Union[str, Tuple[str, str]],
    ) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]

        self._referenced_citations.update(label for label, _ in adapted_labels)

        return Citation(adapted_labels)

    @stateless
    # TODO make this output \citeauthor
    def citeauthor(self, ctx: StatelessContext[MarkdownRenderer], label: str) -> Inline:
        return Citation([(label, None)])


class MarkdownCitationAsHTMLPlugin(Plugin[MarkdownRenderer], CitationPluginInterface):
    _citations: Dict[str, Any]
    _referenced_citations: Set[str]

    def __init__(self) -> None:
        super().__init__()

        # TODO load citations from somewhere
        self._citations = {}
        self._referenced_citations = set()

    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_inline(Citation, self._emit_citation)

    def _postamble_handlers(
        self,
    ) -> Iterable[Tuple[str, Callable[[MarkdownRenderer], None]]]:
        return ((self._BIBLIOGRAPHY_POSTAMBLE_ID, self._emit_bibliography),)

    def _get_citation_shorthand(
        self, label: str, note: Optional[UnescapedText]
    ) -> UnescapedText:
        # TODO could do e.g. numbering here
        if note:
            return UnescapedText(f"[{label}, {note.text}]")
        return UnescapedText(f"[{label}]")

    def _emit_citation(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        citation: Citation,
    ) -> None:
        # TODO what happens with unmarkdownable labels? e.g. labels with backslash or something. need to check that when loading.
        # TODO also maybe people wouldn't want those labels being exposed?

        for label, opt_note in citation.labels:
            renderer.emit(ctx.url(f"#cite-{label}") @ self._get_citation_shorthand(label, opt_note))

    def _emit_bibliography(self, renderer: MarkdownRenderer) -> None:
        # TODO actual reference rendering!
        def emit_individual(label: str) -> None:
            renderer.emit(
                f'<a id="cite-{label}">',
                self._get_citation_shorthand(label, None),
                f" : cite {label}</a>"
            )

        # TODO in HTML mode we should emit these as a list?
        renderer.emit_join(
            emit_individual,
            self._referenced_citations,
            renderer.emit_break_paragraph,
        )


    @stateful
    def cite(
        self,
        state: MutableState[MarkdownRenderer],
        *labels: Union[str, Tuple[str, str]],
    ) -> Inline:
        # Convert ["label"] to [("label", None)] so Citation has a consistent format
        adapted_labels = [
            (label, None)
            if isinstance(label, str)
            else (label[0], UnescapedText(label[1]))
            for label in labels
        ]

        self._referenced_citations.update(label for label, _ in adapted_labels)

        return Citation(adapted_labels)

    # TODO make this output \citeauthor
    @stateful
    def citeauthor(self, state: MutableState[MarkdownRenderer], label: str) -> Inline:
        return Citation([(label, None)])


class MarkdownFootnotePlugin(Plugin[MarkdownRenderer], FootnotePluginInterface):
    _footnotes: Dict[str, Block]
    _footnote_ref_order: List[str]

    _MARKDOWN_FOOTNOTE_POSTAMBLE_ID = "MarkdownFootnotePlugin_Footnotes"

    def __init__(self) -> None:
        super().__init__()

        self._footnotes = {}
        self._footnote_ref_order = []

    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_inline(FootnoteAnchor, self._emit_footnote_anchor)

    def _postamble_handlers(
        self,
    ) -> Iterable[Tuple[str, Callable[[MarkdownRenderer], None]]]:
        return ((self._MARKDOWN_FOOTNOTE_POSTAMBLE_ID, self._emit_footnotes),)

    def _emit_footnote_anchor(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        footnote: FootnoteAnchor,
    ) -> None:
        footnote_num = self._footnote_ref_order.index(footnote.label)
        if renderer.in_html_mode:
            renderer.emit(
                ctx.url(f"#footnote-{footnote_num+1}") @ f"^{footnote_num+1}"
            )
        else:
            renderer.emit_raw(f"[^{footnote_num + 1}]")

    def _emit_footnotes(self, renderer: MarkdownRenderer) -> None:
        def emit_individual(num_label: Tuple[int, str]) -> None:
            num, label = num_label
            if renderer.in_html_mode:
                renderer.emit(
                    f'<a id="footnote-{num+1}">',
                    UnescapedText(f"^{num+1}: "),
                    self._footnotes[label],
                    "</a>"
                )
            else:
                renderer.emit(f"[^{num + 1}]: ", self._footnotes[label])


        return renderer.emit_join(
            emit_individual,
            enumerate(self._footnote_ref_order),
            renderer.emit_break_paragraph,
        )

    def _add_footnote_reference(self, label: str) -> None:
        if label not in self._footnote_ref_order:
            self._footnote_ref_order.append(label)

    @property
    @stateful
    def footnote(self, state: MutableState[MarkdownRenderer]) -> InlineScopeBuilder:
        @inline_scope_builder
        def footnote_builder(contents: InlineScope) -> Inline:
            label = str(uuid.uuid4())
            self._add_footnote_reference(label)
            self._footnotes[label] = Paragraph([Sentence([contents])])
            return FootnoteAnchor(label)

        return footnote_builder

    @stateful
    def footnote_ref(self, state: MutableState[MarkdownRenderer], label: str) -> Inline:
        self._add_footnote_reference(label)
        return FootnoteAnchor(label)

    @stateful
    def footnote_text(
        self, state: MutableState[MarkdownRenderer], label: str
    ) -> BlockScopeBuilder:
        # Return a callable which is invoked with the contents of the following inline scope
        # Example usage:
        # [footnote_text("label")]{text}
        # equivalent to
        # [footnote_text("label")(r"text")]
        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Optional[Block]:
            self._footnotes[label] = contents
            return None

        return handle_block_contents


class MarkdownSectionPlugin(Plugin[MarkdownRenderer], SectionPluginInterface):
    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_block(HeadedBlock, self._emit_headed_block)

    def _emit_headed_block(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        block: HeadedBlock,
    ) -> None:
        if renderer.in_html_mode:
            tag = f"h{block.level}"
            header = f"<{tag}>"
            if block.label:
                header += f'<a id="{block.label}"></a>'
            if block.use_num:
                header += ".".join(str(n) for n in block.num)
                header += " "
            renderer.emit(
                header,
                block.name,
                f"</{tag}>"
            )
            renderer.emit_break_paragraph()
            renderer.emit_blockscope(block.contents)
        else:
            header = "#" * block.level + " "
            if block.label:
                header += f'<a id="{block.label}"></a>'
            if block.use_num:
                header += ".".join(str(n) for n in block.num)
                header += " "
            renderer.emit(
                header,
                block.name
            )
            renderer.emit_break_paragraph()
            renderer.emit_blockscope(block.contents)

    # [section_num, subsection_num... etc]
    _current_number: List[int]

    def __init__(self) -> None:
        super().__init__()
        self._current_number = []

    def _update_number(self, current_level: int) -> Tuple[int, ...]:
        """Update the structure number based on a newly-encountered header at level current_level.

        i.e. if encountering a new section, call _update_number(2), which will a) increment the subsection number and b) return a tuple of (section_num, subsection_num) which can be used in rendering later.
        """

        # Step 1: pad self._current_number out to current_level elements with 0s
        self._current_number += (current_level - len(self._current_number)) * [0]
        # Step 2: increment self._current_number[level - 1] and reset all numbers beyond that to 0
        # e.g. for a plain section, level = 1 => increment current_number[0] and zero out current_number[1:]
        self._current_number[current_level - 1] += 1
        for i in range(current_level, len(self._current_number)):
            self._current_number[i] = 0
        # Step 3: return a tuple of all elements up to current_level
        return tuple(self._current_number[:current_level])

    @stateful
    def section(
        self,
        state: MutableState[MarkdownRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(1)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=1,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents

    @stateful
    def subsection(
        self,
        state: MutableState[MarkdownRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(2)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=2,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents

    @stateful
    def subsubsection(
        self,
        state: MutableState[MarkdownRenderer],
        name: str,
        label: Optional[str] = None,
        num: bool = True,
    ) -> BlockScopeBuilder:
        # Increment the section number here, not at the end of the block
        section_num = self._update_number(3)

        @block_scope_builder
        def handle_block_contents(contents: BlockScope) -> Block:
            return HeadedBlock(
                level=3,
                name=UnescapedText(name),
                label=label,
                use_num=num,
                num=section_num,
                contents=contents,
            )

        return handle_block_contents


@inline_scope_builder
def emph_builder(items: InlineScope) -> Inline:
    return italic_builder.build_from_inlines(items)


@inline_scope_builder
def italic_builder(items: InlineScope) -> Inline:
    # Use single underscore for italics, double asterisks for bold, double underscore for underlining?
    return Formatted("_", "i", items)


@inline_scope_builder
def bold_builder(items: InlineScope) -> Inline:
    return Formatted("**", "b", items)


class MarkdownFormatPlugin(Plugin[MarkdownRenderer], FormatPluginInterface):
    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_inline(Formatted, self._emit_formatted)

    def _emit_formatted(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        item: Formatted,
    ) -> None:
        if renderer.in_html_mode:
            renderer.emit(
                f"<{item.html_format}>",
                item.items,
                f"</{item.html_format}>"
            )
        else:
            renderer.emit(
                item.markdown_format,
                item.items,
                item.markdown_format,
            )

    emph = emph_builder
    italic = italic_builder
    bold = bold_builder

    DQUOTE = RawMarkdown('"')

    @property
    @stateless
    def enquote(self, ctx: StatelessContext[MarkdownRenderer]) -> InlineScopeBuilder:
        @inline_scope_builder
        def enquote_builder(items: InlineScope) -> Inline:
            return InlineScope(
                [MarkdownFormatPlugin.DQUOTE]
                + list(items)
                + [MarkdownFormatPlugin.DQUOTE]
            )

        return enquote_builder


class MarkdownListPlugin(Plugin[MarkdownRenderer]):
    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_block(DisplayList, self._emit_list)

    def _emit_list(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        list: DisplayList,
    ) -> None:
        if renderer.in_html_mode:
            def emit_elem() -> Generator[None, None, None]:
                for item in list.items:
                    renderer.emit_raw("<li>")
                    renderer.emit_newline()
                    with renderer.indent(4):
                        renderer.emit_block(item.contents)
                    renderer.emit_newline()
                    renderer.emit_raw("</li>")
                    yield None
                    
            tag = "ol" if list.numbered else "ul"
            renderer.emit_raw(f"<{tag}>")
            with renderer.indent(4):
                renderer.emit_break_sentence()
                renderer.emit_join_gen(emit_elem(), renderer.emit_break_sentence)
            renderer.emit_raw(f"</{tag}>")
        else:
            if list.numbered:
                def emit_numbered() -> Generator[None, None, None]:
                    for idx, item in enumerate(list.items):
                        indent = f"{idx+1}. "
                        renderer.emit_raw(indent)
                        with renderer.indent(len(indent)):
                            renderer.emit_block(item.contents)
                        yield None

                renderer.emit_join_gen(emit_numbered(), renderer.emit_break_sentence)
            else:
                def emit_dashed() -> Generator[None, None, None]:
                    for idx, item in enumerate(list.items):
                        indent = f"- "
                        renderer.emit_raw(indent)
                        with renderer.indent(len(indent)):
                            renderer.emit_block(item.contents)
                        yield None

                renderer.emit_join_gen(emit_dashed(), renderer.emit_break_sentence)

    @property
    @stateless
    def enumerate(self, ctx: StatelessContext[MarkdownRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def enumerate_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(numbered=True, items=cast(List[DisplayListItem], items))

        return enumerate_builder

    @property
    @stateless
    def itemize(self, ctx: StatelessContext[MarkdownRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def itemize_builder(contents: BlockScope) -> Block:
            items = list(contents)
            if any(not isinstance(x, DisplayListItem) for x in items):
                raise TypeError(
                    f"Found blocks in this list that were not list [item]s!"
                )
            return DisplayList(numbered=False, items=cast(List[DisplayListItem], items))

        return itemize_builder

    @property
    @stateless
    def item(self, ctx: StatelessContext[MarkdownRenderer]) -> BlockScopeBuilder:
        @block_scope_builder
        def item_builder(block_scope: BlockScope) -> Block:
            return DisplayListItem(block_scope)

        return item_builder


class MarkdownUrlPlugin(Plugin[MarkdownRenderer]):
    def _add_emitters(self, handler: CustomEmitDispatch[MarkdownRenderer]) -> None:
        handler.add_custom_inline(NamedUrl, self._emit_url)

    def _emit_url(
        self,
        renderer: MarkdownRenderer,
        ctx: StatelessContext[MarkdownRenderer],
        url: NamedUrl,
    ) -> None:
        if renderer.in_html_mode:
            assert ">" not in url.url and "<" not in url.url
            renderer.emit_raw(f'<a href="{url.url}">')
            if url.name is None:
                # Set the "name" of the URL to the text of the URL - escaped so it can be read as normal
                renderer.emit_unescapedtext(UnescapedText(url.url))
            else:
                renderer.emit_inlinescope(url.name)
            renderer.emit_raw("</a>")
        else:
            assert ")" not in url.url
            renderer.emit_raw("[")
            if url.name is None:
                # Set the "name" of the URL to the text of the URL - escaped so it can be read as normal markdown
                renderer.emit_unescapedtext(UnescapedText(url.url))
            else:
                renderer.emit_inlinescope(url.name)
            renderer.emit_raw(f"]({url.url})")

    @stateless
    def url(
        self,
        ctx: StatelessContext[MarkdownRenderer],
        url: str,
        name: Optional[str] = None,
    ) -> Inline:
        return NamedUrl(
            url, name=InlineScope([UnescapedText(name)]) if name is not None else None
        )



from typing import Any, Callable, List, Optional, Union

from turnip_text import (
    Block,
    BlockScope,
    BlockScopeBuilder,
    CoercibleToBlockScope,
    CoercibleToInlineScope,
    Inline,
    InlineScope,
    InlineScopeBuilder,
    RawScopeBuilder,
    coerce_to_block_scope,
    coerce_to_inline_scope,
)

# TODO tests for the helpers


class block_scope_builder(BlockScopeBuilder):
    """
    Decorator which allows a function to fit the BlockScopeBuilder typeclass.

    e.g. one could define a function
    ```python
    def block(name=""):
        @block_scope_builder
        def inner(items: BlockScope) -> Block:
            return items
        return inner
    ```
    which allows turnip-text as so:
    ```!text
    [block(name="greg")]{
    The contents of greg
    }
    ```

    It also supports the matmul operator, which tries to coerce the right-hand-side into a BlockScope before calling the function:
    ```python
    block(name="greg") @ "The contents of greg"
    ```
    """

    func: Callable[[BlockScope], Optional[Block]]

    def __init__(self, func: Callable[[BlockScope], Optional[Block]]) -> None:
        self.func = func

    def build_from_blocks(self, b: BlockScope) -> Optional[Block]:
        return self.func(b)


class inline_scope_builder(InlineScopeBuilder):
    """
    Decorator which allows a function to fit the InlineScopeBuilder typeclass.

    e.g. one could define a function
    ```python
    def inline(postfix = ""):
        @inline_scope_builder
        def inner(items: InlineScope) -> Inline:
            return InlineScope(list(items) + [postfix])
        return inner
    ```
    which allows turnip-text as so:
    ```!text
    [inline("!")]{surprise}
    ```

    It also supports the matmul operator, which tries to coerce the right-hand-side into an InlineScope before calling the function:
    ```python
    inline("!") @ "surprise"
    ```
    """

    func: Callable[[InlineScope], Inline]

    def __init__(self, func: Callable[[InlineScope], Inline]) -> None:
        self.func = func

    def build_from_inlines(self, inls: InlineScope) -> Inline:
        return self.func(inls)


class raw_scope_builder(RawScopeBuilder):
    """
    Decorator which allows a function to fit the RawScopeBuilder typeclass.

    e.g. one could define a function
    ```python
    @raw_scope_builder
    def math(raw_text: str) -> Inline:
        ...
    ```
    which allows turnip-text as so:
    ```!text
    [math]#{\\sin x}#
    ```

    It also supports the matmul operator, which checks the right-hand-side is a str before calling the function:
    ```python
    math @ r"\\sin x"
    ```
    """

    func: Callable[[str], Union[Block, Inline]]

    def __init__(self, func: Callable[[str], Union[Block, Inline]]) -> None:
        self.func = func

    def build_from_raw(self, raw: str) -> Union[Block, Inline]:
        return self.func(raw)

    def __matmul__(self, maybe_str: Any) -> Union[Block, Inline]:
        if isinstance(maybe_str, str):
            return self.func(maybe_str)
        raise TypeError(
            f"Invoked RawScopeBuilder on {maybe_str}, which wasn't a string"
        )



from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Callable, Iterator, List

from turnip_text import Inline, UnescapedText
from turnip_text.renderers import Plugin, Renderer


@dataclass(frozen=True)
class RawLatex(Inline):
    text: str


class LatexRenderer(Renderer):
    def __init__(self, plugins: List[Plugin["LatexRenderer"]]) -> None:
        super().__init__(plugins)

        self.emit_dispatch.add_custom_inline(
            RawLatex, lambda r, ctx, raw: r._emit_raw_latex(raw)
        )

    def _emit_raw_latex(self, r: RawLatex) -> None:
        self.emit_raw(r.text)

    def emit_unescapedtext(self, t: UnescapedText) -> None:
        # note - right now this assumes we're using a unicode-compatible setup and thus don't need to escape unicode characters.
        # note - order is important here because the subsitutions may introduce more special characters.
        # e.g. if the backslash replacement applied later, the backslash from escaping "%" would be replaced with \textbackslash
        ascii_map = {
            "\\": "\\textbackslash{}",
            "%": "\%",
            "$": "\$",
            "{": "\{",
            "}": "\}",
            "_": "\_",
            "#": "\#",
            "&": "\&",
            "~": "\~{}",
        }
        data = t.text
        for c, replace_with in ascii_map.items():
            data = data.replace(c, replace_with)
        self.emit_raw(data)

    def emit_macro(self, name: str) -> None:
        self.emit_raw(f"\\{name}")

    def emit_sqr_bracketed(self, *args: Any) -> None:
        self.emit_raw("[")
        self.emit(*args)
        self.emit_raw("]")

    def emit_braced(self, *args: Any) -> None:
        self.emit_raw("{")
        self.emit(*args)
        self.emit_raw("}")

    @contextmanager
    def emit_env(self, name: str, indent: int = 4) -> Iterator[None]:
        self.emit_macro("begin")
        self.emit_braced(name)
        self.push_indent(indent)
        self.emit_break_sentence()

        try:
            yield
        finally:
            self.pop_indent(indent)
            self.emit_break_sentence()
            self.emit_macro("end")
            self.emit_braced(name)
    